@@ -2,9 +2,8 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.pool import Pool
-from . import stock
-from . import product
 
+from . import product, stock
 from .stock import LotMixin
 
 __all__ = ['LotMixin', 'register']
@@ -13,6 +12,7 @@ __all__ = ['LotMixin', 'register']
 def register():
     Pool.register(
         stock.Lot,
+        stock.LotTrace,
         stock.LotByLocationContext,
         stock.LotsByLocations,
         stock.LotByWarehouseContext,
@@ -38,3 +38,7 @@ def register():
         stock.MoveAddLots,
         stock.InventoryCount,
         module='stock_lot', type_='wizard')
+    Pool.register(
+        stock.ShipmentDrop,
+        module='stock_lot', type_='model',
+        depends=['sale_supply_drop_shipment'])
