@@ -1,21 +1,25 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import datetime
+from collections import defaultdict
 from copy import copy
 from functools import wraps
 
 from sql import Column
+from sql.functions import CharLength
 
 from trytond.i18n import gettext
-from trytond.model import Model, ModelView, ModelSQL, fields
+from trytond.model import (
+    DeactivableMixin, Index, Model, ModelSQL, ModelView, fields)
 from trytond.model.exceptions import (
     AccessError, RequiredValidationError, ValidationError)
-from trytond.pyson import Eval, Bool, Len
+from trytond.modules.stock import StockMixin
+from trytond.modules.stock.exceptions import ShipmentCheckQuantityWarning
 from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, Len
 from trytond.tools import grouped_slice
 from trytond.transaction import Transaction
-from trytond.modules.stock import StockMixin
-from trytond.wizard import Wizard, StateView, StateTransition, Button
+from trytond.wizard import Button, StateTransition, StateView, Wizard
 
 
 def check_no_move(func):
@@ -36,8 +40,7 @@ def check_no_move(func):
     @wraps(func)
     def decorator(cls, *args):
         transaction = Transaction()
-        if (transaction.user != 0
-                and transaction.context.get('_check_access')):
+        if transaction.user and transaction.check_access:
             actions = iter(args)
             for records, values in zip(actions, actions):
                 for field, state, error in cls._modify_no_move:
@@ -51,37 +54,43 @@ def check_no_move(func):
 
 
 class LotMixin:
-
+    __slots__ = ()
     number = fields.Char(
-        "Number", required=True, select=True,
+        "Number", required=True,
         states={
             'required': ~Eval('has_sequence') | (Eval('id', -1) >= 0),
-            },
-        depends=['has_sequence'])
+            })
     product = fields.Many2One('product.product', 'Product', required=True)
     has_sequence = fields.Function(
         fields.Boolean("Has Sequence"), 'on_change_with_has_sequence')
 
+    @classmethod
+    def __setup__(cls):
+        cls.number.search_unaccented = False
+        super().__setup__()
+
     @fields.depends('product')
     def on_change_with_has_sequence(self, name=None):
         if self.product:
             return bool(self.product.lot_sequence)
 
 
-class Lot(ModelSQL, ModelView, LotMixin, StockMixin):
+class Lot(DeactivableMixin, ModelSQL, ModelView, LotMixin, StockMixin):
     "Stock Lot"
     __name__ = 'stock.lot'
     _rec_name = 'number'
 
-    quantity = fields.Function(fields.Float('Quantity'), 'get_quantity',
-        searcher='search_quantity')
-    forecast_quantity = fields.Function(fields.Float('Forecast Quantity'),
+    quantity = fields.Function(
+        fields.Float("Quantity", digits='default_uom'),
+        'get_quantity', searcher='search_quantity')
+    forecast_quantity = fields.Function(
+        fields.Float("Forecast Quantity", digits='default_uom'),
         'get_quantity', searcher='search_quantity')
     default_uom = fields.Function(
-        fields.Many2One('product.uom', "Default UOM"),
+        fields.Many2One(
+            'product.uom', "Default UoM",
+            help="The default Unit of Measure."),
         'on_change_with_default_uom', searcher='search_default_uom')
-    default_uom_digits = fields.Function(fields.Integer("Default Unit Digits"),
-        'on_change_with_default_uom_digits')
 
     @classmethod
     def __setup__(cls):
@@ -89,13 +98,26 @@ class Lot(ModelSQL, ModelView, LotMixin, StockMixin):
         cls._modify_no_move = [
             ('product', None, 'stock_lot.msg_change_product'),
             ]
+        cls._buttons.update({
+                'upward_traces': {},
+                'downward_traces': {},
+                })
+
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
 
     @classmethod
     def get_quantity(cls, lots, name):
         location_ids = Transaction().context.get('locations')
         product_ids = list(set(l.product.id for l in lots))
-        return cls._get_quantity(lots, name, location_ids,
-            grouping=('product', 'lot',), grouping_filter=(product_ids,))
+        quantities = {}
+        for product_ids in grouped_slice(product_ids):
+            quantities.update(cls._get_quantity(lots, name, location_ids,
+                    grouping=('product', 'lot',),
+                    grouping_filter=(list(product_ids),)))
+        return quantities
 
     @classmethod
     def search_quantity(cls, name, domain=None):
@@ -105,18 +127,20 @@ class Lot(ModelSQL, ModelView, LotMixin, StockMixin):
 
     @fields.depends('product')
     def on_change_with_default_uom(self, name=None):
-        if self.product:
-            return self.product.default_uom.id
+        return self.product.default_uom if self.product else None
 
     @classmethod
     def search_default_uom(cls, name, clause):
         nested = clause[0][len(name):]
         return [('product.' + name + nested, *clause[1:])]
 
-    @fields.depends('product')
-    def on_change_with_default_uom_digits(self, name=None):
-        if self.product:
-            return self.product.default_uom.digits
+    @classmethod
+    def copy(cls, lots, default=None):
+        default = default.copy() if default else {}
+        has_sequence = {l.id: l.has_sequence for l in lots}
+        default.setdefault(
+            'number', lambda o: None if has_sequence[o['id']] else o['number'])
+        return super().copy(lots, default=default)
 
     @classmethod
     def create(cls, vlist):
@@ -140,6 +164,120 @@ class Lot(ModelSQL, ModelView, LotMixin, StockMixin):
     def write(cls, *args):
         super(Lot, cls).write(*args)
 
+    @classmethod
+    @ModelView.button_action('stock_lot.act_lot_trace_upward_relate')
+    def upward_traces(cls, lots):
+        pass
+
+    @classmethod
+    @ModelView.button_action('stock_lot.act_lot_trace_downward_relate')
+    def downward_traces(cls, lots):
+        pass
+
+
+class LotTrace(ModelSQL, ModelView):
+    "Lot Trace"
+    __name__ = 'stock.lot.trace'
+
+    product = fields.Many2One(
+        'product.product', "Product",
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
+    lot = fields.Many2One('stock.lot', "Lot")
+
+    quantity = fields.Float("Quantity", digits='unit')
+    unit = fields.Many2One('product.uom', "Unit")
+
+    company = fields.Many2One('company.company', "Company")
+    date = fields.Date("Date")
+    shipment = fields.Reference("Shipment", 'get_shipments')
+    document = fields.Function(
+        fields.Reference("Document", 'get_documents'), 'get_document')
+
+    upward_traces = fields.Function(
+        fields.Many2Many(
+            'stock.lot.trace', None, None, "Upward Traces"),
+        'get_upward_traces')
+    downward_traces = fields.Function(
+        fields.Many2Many(
+            'stock.lot.trace', None, None, "Downward Traces"),
+        'get_downward_traces')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._order.insert(0, ('date', 'ASC'))
+
+    @classmethod
+    def table_query(cls):
+        pool = Pool()
+        Move = pool.get('stock.move')
+        move = Move.__table__()
+        return (
+            move.select(
+                *cls._columns(move),
+                where=move.state == 'done'))
+
+    @classmethod
+    def _columns(cls, move):
+        return [
+            move.id.as_('id'),
+            move.create_uid.as_('create_uid'),
+            move.create_date.as_('create_date'),
+            move.write_uid.as_('write_uid'),
+            move.write_date.as_('write_date'),
+            move.product.as_('product'),
+            move.lot.as_('lot'),
+            move.quantity.as_('quantity'),
+            move.unit.as_('unit'),
+            move.company.as_('company'),
+            move.effective_date.as_('date'),
+            move.shipment.as_('shipment'),
+            ]
+
+    def get_rec_name(self, name):
+        return self.document.rec_name if self.document else str(self.id)
+
+    @classmethod
+    def get_shipments(cls):
+        pool = Pool()
+        Move = pool.get('stock.move')
+        return Move.get_shipment()
+
+    @classmethod
+    def get_documents(cls):
+        return cls.get_shipments()
+
+    def get_document(self, name):
+        if self.shipment:
+            return str(self.shipment)
+
+    @classmethod
+    def _is_trace_move(cls, move):
+        return move.state == 'done'
+
+    @classmethod
+    def _trace_move_order_key(cls, move):
+        return (move.effective_date,)
+
+    def get_upward_traces(self, name):
+        return list(map(int, sorted(filter(
+                        self._is_trace_move, self._get_upward_traces()),
+                    key=self._trace_move_order_key)))
+
+    def _get_upward_traces(self):
+        return set()
+
+    def get_downward_traces(self, name):
+        return list(map(int, sorted(filter(
+                        self._is_trace_move, self._get_downward_traces()),
+                    key=self._trace_move_order_key)))
+
+    def _get_downward_traces(self):
+        return set()
+
 
 class LotByLocationContext(ModelView):
     'Lot by Location'
@@ -171,20 +309,16 @@ class LotsByLocations(ModelSQL, ModelView):
     lot = fields.Many2One('stock.lot', "Lot")
     product = fields.Many2One('product.product', "Product")
     quantity = fields.Function(
-        fields.Float(
-            "Quantity", digits=(16, Eval('default_uom_digits', 2)),
-            depends=['default_uom_digits']),
+        fields.Float("Quantity", digits='default_uom'),
         'get_lot', searcher='search_lot')
     forecast_quantity = fields.Function(
-        fields.Float(
-            "Forecast Quantity", digits=(16, Eval('default_uom_digits', 2)),
-            depends=['default_uom_digits']),
+        fields.Float("Forecast Quantity", digits='default_uom'),
         'get_lot', searcher='search_lot')
     default_uom = fields.Function(
-        fields.Many2One('product.uom', "Default UOM"),
+        fields.Many2One(
+            'product.uom', "Default UoM",
+            help="The default Unit of Measure."),
         'get_lot', searcher='search_lot')
-    default_uom_digits = fields.Function(
-        fields.Integer("Default UOM Digits"), 'get_lot')
 
     @classmethod
     def __setup__(cls):
@@ -208,6 +342,13 @@ class LotsByLocations(ModelSQL, ModelView):
                 columns.append(column.as_(fname))
         return lot.select(*columns)
 
+    def get_rec_name(self, name):
+        return self.lot.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return [('lot.rec_name',) + tuple(clause[1:])]
+
     def get_lot(self, name):
         value = getattr(self.lot, name)
         if isinstance(value, Model):
@@ -216,8 +357,8 @@ class LotsByLocations(ModelSQL, ModelView):
 
     @classmethod
     def search_lot(cls, name, clause):
-        nested = clause[0].lstrip(name)
-        return [('lot.' + name + nested,) + tuple(clause[1:])]
+        nested = clause[0][len(name):]
+        return [('lot.' + name + nested, *clause[1:])]
 
 
 class LotByWarehouseContext(LotByLocationContext):
@@ -229,6 +370,9 @@ class LotByWarehouseContext(LotByLocationContext):
             ('type', '=', 'warehouse'),
             ],
         )
+    stock_skip_warehouse = fields.Boolean(
+        "Only storage zone",
+        help="Check to use only the quantity of the storage zone.")
     locations = fields.Function(
         fields.Many2Many('stock.location', None, None, "Locations"),
         'on_change_with_locations')
@@ -237,11 +381,15 @@ class LotByWarehouseContext(LotByLocationContext):
     def default_warehouse(cls):
         return Pool().get('stock.location').get_default_warehouse()
 
+    @classmethod
+    def default_stock_skip_warehouse(cls):
+        return Transaction().context.get('stock_skip_warehouse')
+
     @fields.depends('warehouse')
     def on_change_with_locations(self, name=None):
         locations = []
         if self.warehouse:
-            locations.append(self.warehouse.id)
+            locations.append(self.warehouse)
         return locations
 
 
@@ -274,8 +422,7 @@ class Move(metaclass=PoolMeta):
             ],
         states={
             'readonly': Eval('state').in_(['cancelled', 'done']),
-            },
-        depends=['state', 'product'])
+            })
 
     @classmethod
     def __setup__(cls):
@@ -293,6 +440,12 @@ class Move(metaclass=PoolMeta):
     def add_lots_wizard(cls, moves):
         pass
 
+    def add_lot(self):
+        if not self.lot and self.product:
+            lot = self.product.create_lot()
+            if lot:
+                self.lot = lot
+
     def check_lot(self):
         "Check if lot is required"
         if (self.internal_quantity
@@ -310,6 +463,23 @@ class Move(metaclass=PoolMeta):
         for move in moves:
             move.check_lot()
 
+    @classmethod
+    def assign_try(cls, moves, with_childs=True, grouping=('product',)):
+        if 'lot' not in grouping:
+            moves_with_lot, moves_without_lot = [], []
+            for move in moves:
+                if move.lot:
+                    moves_with_lot.append(move)
+                else:
+                    moves_without_lot.append(move)
+            success = super().assign_try(
+                moves_with_lot, with_childs, grouping=grouping + ('lot',))
+            success &= super().assign_try(
+                moves_without_lot, with_childs, grouping=grouping)
+        else:
+            success = super().assign_try(moves, with_childs, grouping)
+        return success
+
 
 class MoveAddLots(Wizard):
     "Add Lots"
@@ -328,7 +498,7 @@ class MoveAddLots(Wizard):
         if 'quantity' in fields:
             default['quantity'] = self.record.quantity
         if 'unit' in fields:
-            default['unit'] = self.record.uom.id
+            default['unit'] = self.record.unit.id
         return default
 
     def transition_add(self):
@@ -338,14 +508,13 @@ class MoveAddLots(Wizard):
         lang = Lang.get()
         quantity_remaining = self.start.on_change_with_quantity_remaining()
         if quantity_remaining < 0:
-            digits = self.record.uom.digits
+            digits = self.record.unit.digits
             move_quantity = self.record.quantity
             lot_quantity = self.record.quantity - quantity_remaining
             raise ValidationError(gettext(
                     'stock_lot.msg_move_add_lot_quantity',
-                    lot_quantity=lang.format('%.*f', (digits, lot_quantity)),
-                    move_quantity=lang.format(
-                        '%.*f', (digits, move_quantity))))
+                    lot_quantity=lang.format_number(lot_quantity, digits),
+                    move_quantity=lang.format_number(move_quantity, digits)))
         lots = []
         for line in self.start.lots:
             lot = line.get_lot(self.record)
@@ -373,16 +542,10 @@ class MoveAddLotsStart(ModelView):
     __name__ = 'stock.move.add.lots.start'
 
     product = fields.Many2One('product.product', "Product", readonly=True)
-    quantity = fields.Float(
-        "Quantity", digits=(16, Eval('unit_digits', 2)), readonly=True,
-        depends=['unit_digits'])
+    quantity = fields.Float("Quantity", digits='unit', readonly=True)
     unit = fields.Many2One('product.uom', "Unit", readonly=True)
-    unit_digits = fields.Function(
-        fields.Integer("Unit Digits"), 'on_change_with_unit_digits')
     quantity_remaining = fields.Function(
-        fields.Float(
-            "Quantity Remaining", digits=(16, Eval('unit_digits', 2)),
-            depends=['unit_digits']),
+        fields.Float("Quantity Remaining", digits='unit'),
         'on_change_with_quantity_remaining')
 
     lots = fields.One2Many(
@@ -392,15 +555,13 @@ class MoveAddLotsStart(ModelView):
             ],
         states={
             'readonly': ~Eval('quantity_remaining', 0),
-            },
-        depends=['product', 'quantity_remaining'])
+            })
 
     duplicate_lot_number = fields.Integer(
         "Duplicate Lot Number",
         states={
             'invisible': Len(Eval('lots')) != 1,
             },
-        depends=['lots'],
         help="The number of times the lot must be duplicated.")
 
     @classmethod
@@ -415,11 +576,6 @@ class MoveAddLotsStart(ModelView):
                 },
             )
 
-    @fields.depends('unit')
-    def on_change_with_unit_digits(self, name=None):
-        if self.unit:
-            return self.unit.digits
-
     @fields.depends('quantity', 'lots', 'unit')
     def on_change_with_quantity_remaining(self, name=None):
         if self.quantity is not None:
@@ -450,11 +606,9 @@ class MoveAddLotsStartLot(ModelView, LotMixin):
     __name__ = 'stock.move.add.lots.start.lot'
 
     parent = fields.Many2One('stock.move.add.lots.start', "Parent")
-    quantity = fields.Float(
-        "Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
-        depends=['unit_digits'])
-    unit_digits = fields.Function(
-        fields.Integer("Unit Digits"), 'on_change_with_unit_digits')
+    quantity = fields.Float("Quantity", digits='quantity_unit', required=True)
+    quantity_unit = fields.Function(
+        fields.Many2One('product.uom', "Unit"), 'on_change_with_quantity_unit')
 
     @fields.depends(
         'parent', '_parent_parent.quantity_remaining')
@@ -463,10 +617,10 @@ class MoveAddLotsStartLot(ModelView, LotMixin):
                 and self.parent.quantity_remaining is not None):
             self.quantity = self.parent.quantity_remaining
 
-    @fields.depends('parent', '_parent_parent.unit_digits')
-    def on_change_with_unit_digits(self, name=None):
-        if self.parent:
-            return self.parent.unit_digits
+    @fields.depends('parent', '_parent_parent.unit')
+    def on_change_with_quantity_unit(self, name=None):
+        if self.parent and self.parent.unit:
+            return self.parent.unit.id
 
     @fields.depends('number', 'product', methods=['_set_lot_values'])
     def on_change_number(self):
@@ -506,7 +660,59 @@ class MoveAddLotsStartLot(ModelView, LotMixin):
             }
 
 
-class ShipmentIn(metaclass=PoolMeta):
+class ShipmentCheckQuantity:
+    "Check quantities per lot between source and target moves"
+    __slots__ = ()
+
+    def check_quantity(self):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        Lang = pool.get('ir.lang')
+        lang = Lang.get()
+
+        super().check_quantity()
+
+        products_with_lot = set()
+        source_qties = defaultdict(float)
+        for move in self._check_quantity_source_moves:
+            if move.lot:
+                products_with_lot.add(move.product)
+                source_qties[move.lot] += move.internal_quantity
+
+        target_qties = defaultdict(float)
+        for move in self._check_quantity_target_moves:
+            if move.lot:
+                target_qties[move.lot] += move.internal_quantity
+
+        diffs = {}
+        for lot, incoming_qty in target_qties.items():
+            if (lot not in source_qties
+                    and lot.product not in products_with_lot):
+                continue
+            unit = lot.product.default_uom
+            incoming_qty = unit.round(incoming_qty)
+            inventory_qty = unit.round(source_qties.pop(lot, 0))
+            diff = inventory_qty - incoming_qty
+            if diff:
+                diffs[lot] = diff
+
+        if diffs:
+            warning_name = Warning.format(
+                'check_quantity_lot', [self])
+            if Warning.check(warning_name):
+                quantities = []
+                for lot, quantity in diffs.items():
+                    quantity = lang.format_number_symbol(
+                        quantity, lot.product.default_uom)
+                    quantities.append(f"{lot.rec_name}: {quantity}")
+                raise ShipmentCheckQuantityWarning(warning_name,
+                    gettext(
+                        'stock.msg_shipment_check_quantity',
+                        shipment=self.rec_name,
+                        quantities=', '.join(quantities)))
+
+
+class ShipmentIn(ShipmentCheckQuantity, metaclass=PoolMeta):
     __name__ = 'stock.shipment.in'
 
     def _get_inventory_move(self, incoming_move):
@@ -516,7 +722,7 @@ class ShipmentIn(metaclass=PoolMeta):
         return move
 
 
-class ShipmentOut(metaclass=PoolMeta):
+class ShipmentOut(ShipmentCheckQuantity, metaclass=PoolMeta):
     __name__ = 'stock.shipment.out'
 
     def _get_inventory_move(self, outgoing_move):
@@ -529,7 +735,7 @@ class ShipmentOut(metaclass=PoolMeta):
         return super()._sync_move_key(move) + (('lot', move.lot),)
 
 
-class ShipmentOutReturn(metaclass=PoolMeta):
+class ShipmentOutReturn(ShipmentCheckQuantity, metaclass=PoolMeta):
     __name__ = 'stock.shipment.out.return'
 
     def _get_inventory_move(self, incoming_move):
@@ -539,13 +745,20 @@ class ShipmentOutReturn(metaclass=PoolMeta):
         return move
 
 
-class ShipmentInternal(metaclass=PoolMeta):
+class ShipmentInternal(ShipmentCheckQuantity, metaclass=PoolMeta):
     __name__ = 'stock.shipment.internal'
 
     def _sync_move_key(self, move):
         return super()._sync_move_key(move) + (('lot', move.lot),)
 
 
+class ShipmentDrop(ShipmentCheckQuantity, metaclass=PoolMeta):
+    __name__ = 'stock.shipment.drop'
+
+    def _sync_move_key(self, move):
+        return super()._sync_move_key(move) + (('lot', move.lot),)
+
+
 class Period(metaclass=PoolMeta):
     __name__ = 'stock.period'
     lot_caches = fields.One2Many('stock.period.cache.lot', 'period',
@@ -571,16 +784,34 @@ class PeriodCacheLot(ModelSQL, ModelView):
     It is used to store cached computation of stock quantities per lot.
     '''
     __name__ = 'stock.period.cache.lot'
-    period = fields.Many2One('stock.period', 'Period', required=True,
-        readonly=True, select=True, ondelete='CASCADE')
-    location = fields.Many2One('stock.location', 'Location', required=True,
-        readonly=True, select=True, ondelete='CASCADE')
+    period = fields.Many2One(
+        'stock.period', "Period",
+        required=True, readonly=True, ondelete='CASCADE')
+    location = fields.Many2One(
+        'stock.location', "Location",
+        required=True, readonly=True, ondelete='CASCADE')
     product = fields.Many2One('product.product', 'Product', required=True,
         readonly=True, ondelete='CASCADE')
     lot = fields.Many2One('stock.lot', 'Lot', readonly=True,
         ondelete='CASCADE')
     internal_quantity = fields.Float('Internal Quantity', readonly=True)
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(
+                    t,
+                    (t.period, Index.Equality()),
+                    (t.product, Index.Equality()),
+                    (t.lot, Index.Equality()),
+                    include=[t.internal_quantity]),
+                Index(
+                    t,
+                    (t.location, Index.Equality())),
+                })
+
 
 class Inventory(metaclass=PoolMeta):
     __name__ = 'stock.inventory'
@@ -598,8 +829,7 @@ class InventoryLine(metaclass=PoolMeta):
             ],
         states={
             'readonly': Eval('inventory_state') != 'draft',
-            },
-        depends=['product', 'inventory_state'])
+            })
 
     @classmethod
     def __setup__(cls):
